from __future__ import absolute_import

import logging
from itertools import takewhile
from six.moves import zip_longest

from dcard.api import api, route
from dcard.manager import ContentParser, Downloader
from dcard.utils import flatten_lists

logger = logging.getLogger(__name__)


class Post:

    comments_per_page = 30

    def __init__(self, metadata=None):
        self.only_id = False
        self.ids = []
        self.metas = None
        self._initial_metadata(metadata)

    def __call__(self, value):
        self._initial_metadata(value)
        return self

    def get(self, content=True, links=True, comments=True):

        _content = self.gen_content_reqs(self.ids) if content else []
        _links = self.gen_links_reqs(self.ids) if links else []
        _comments = self.get_comments(self.ids, self.metas) if comments else ()

        def gen_posts():
            for content, links, comments in zip_longest(
                api.imap(_content), api.imap(_links), _comments
            ):
                post = {}
                post.update(content.json()) if content else None
                post.update({
                    'links': links.json() if links else None,
                    'comments': self.extract_comments(comments)
                })
                if post:
                    yield post

            logger.info('[Posts.gen_posts <gen>] Processed.')

        return PostsResult(gen_posts)

    def gen_content_reqs(self, post_ids):
        return (api.get_post(post_id) for post_id in post_ids)

    def gen_links_reqs(self, post_ids):
        return (api.get_post_links(post_id) for post_id in post_ids)

    def extract_comments(self, comments):
        return flatten_lists([cs.json() for cs in api.imap(comments)]) \
            if not self.only_id and comments else comments

    def get_comments(self, post_ids, post_metas):
        return (
            self.get_comments_parallel(meta['id'], meta['commentCount'])
            for meta in post_metas
        ) if post_metas else (
            self.get_comments_serial(post_id)
            for post_id in post_ids
        )

    def get_comments_parallel(self, post_id, comments_count):
        pages = -(-comments_count // self.comments_per_page)
        reqs = (
            api.get_post(
                post_id, 'comments',
                params={'after': page * self.comments_per_page})
            for page in range(pages)
        )
        return reqs

    def get_comments_serial(self, post_id):
        params = {}

        def gen_cmts():
            while True:
                yield api.get_json(route.post_comments(post_id), params=params)

        comments = []
        for cmts in takewhile(lambda x: len(x), gen_cmts()):
            comments += cmts
            params['after'] = cmts[-1]['floor']

        return comments

    def _initial_metadata(self, metadata):
        if not metadata:
            return
        metadata = metadata if isinstance(metadata, list) else [metadata]
        self.only_id = type(metadata[0]) is int
        self.ids = metadata if self.only_id else [m['id'] for m in metadata]
        self.metas = None if self.only_id else metadata


class PostsResult:

    downloader = Downloader()
    parser = ContentParser()

    def __init__(self, generator):
        logger.info('[PostResult] takes hand.')
        self.results = generator()

    def result(self):
        self.results = list(self.results)
        return self.results

    def parse_resources(self):
        self.parser.posts = self.results
        return self.parser.parse()

    def download(self, resource_bundles):
        self.downloader.resource_bundles = resource_bundles
        return self.downloader.download()



from __future__ import absolute_import

import logging
from logging import NullHandler

from dcard.dcard import Dcard  # noqa


__version__ = '0.3.0'
__author__ = 'Salas leVirve'


logging.getLogger(__name__).addHandler(NullHandler())


def add_handles_on_logger():
    logger = logging.getLogger(__name__)
    logger.setLevel(logging.DEBUG)

    fh = logging.FileHandler('dcard.log', encoding='utf8')
    ch = logging.StreamHandler()
    fh.setLevel(logging.DEBUG)
    ch.setLevel(logging.WARNING)

    fh_formatter = logging.Formatter(
        '%(asctime)s - %(name)s - %(levelname)s - %(message)s')
    ch_formatter = logging.Formatter(
        '%(asctime)s - %(message)s')
    fh.setFormatter(fh_formatter)
    ch.setFormatter(ch_formatter)
    logger.addHandler(fh)
    logger.addHandler(ch)



# -*- coding: utf-8 -*-

import logging
import itertools
from multiprocessing.dummy import Pool
from six.moves import http_client as httplib

import requests
from requests.adapters import HTTPAdapter
from requests.packages.urllib3.util.retry import Retry
from requests.exceptions import RetryError

from . import prequests


logger = logging.getLogger(__name__)


class Client:

    max_retries = 5

    def __init__(self, workers=8):
        retries = Retry(
            total=self.max_retries,
            backoff_factor=0.1,
            status_forcelist=[500, 502, 503, 504])
        session = requests.Session()
        session.mount('https://', HTTPAdapter(max_retries=retries))
        self.session = session
        self.pool = Pool(workers)

    def get_json(self, url, **kwargs):
        response = None
        try:
            response = self.session.get(url, **kwargs)
            data = response.json()
            if type(data) is dict and data.get('error'):
                raise ServerResponsedError
            return data
        except ValueError as e:
            retries = kwargs.get('retries', 0)
            logger.error('when get <%d> %s, error %s (retry#%d)',
                         response.status_code, url, e, retries)
            return {} if retries <= self.max_retries else \
                self.get_json(url, retries=retries + 1)
        except ServerResponsedError:
            logger.error('when get <%d> %s, response: %s',
                         response.status_code, url, data)
            return {}
        except httplib.IncompleteRead as e:
            logger.error('when get %s, error %s; partial: %s',
                         url, e, e.partial)
            return {}  # or shall we return `e.partial` ?
        except RetryError as e:
            logger.error('when get %s, retry error occurs. %s', url, e)
            return {}
        except Exception as e:
            logger.error('error %s', e)
            return {}

    def get_stream(self, url, **kwargs):
        request = self.session.get(url, stream=True, **kwargs)
        return request

    def get(self, url, **kwargs):
        return prequests.get(url, session=self.session, **kwargs)

    def imap(self, reqs):
        return prequests.imap(reqs, stream=False, pool=self.pool)


def flatten_lists(meta_lists):
    return list(itertools.chain.from_iterable(meta_lists))


class ServerResponsedError(Exception):
    pass



# -*- coding: utf-8 -*-
from __future__ import unicode_literals, absolute_import
import logging
from itertools import takewhile, count
try:
    from functools import partialmethod
except ImportError:
    from functools import partial

    class partialmethod(partial):
        def __get__(self, instance, owner):
            if instance is None:
                return self
            return partial(
                self.func, instance,
                *(self.args or ()), **(self.keywords or {}))

from six.moves import zip

from dcard.utils import Client

logger = logging.getLogger(__name__)


class Api():

    metas_per_page = 30
    comments_per_page = 30

    def __init__(self, workers=8):
        self.client = Client(workers=workers)

    def get_all_forums(self):
        return self.client.get_json(route.forums())

    def get_general_forums(self):
        forums = self.client.get_json(route.forums())
        return [forum for forum in forums if not forum['isSchool']]

    def get_metas(self, name, sort, num, before, timebound=''):

        def filter_metas(metas):
            if num >= 0 and page == pages:
                metas = metas[:num - (pages - 1) * self.metas_per_page]
            if timebound:
                metas = [m for m in metas if m['updatedAt'] > timebound]
            return metas

        def eager_for_metas(bundle):
            page, metas = bundle
            if num >= 0 and page == pages + 1:
                return False
            if len(metas) == 0:
                logger.warning('[%s] 已到最末頁，第%d頁!', name, page)
            return len(metas) != 0

        def get_single_page_metas():
            while True:
                yield self.client.get_json(url, params=params)

        url = route.posts_meta(name)
        params = {'popular': 'true' if sort == 'popular' else 'false'}
        if before:
            params['before'] = before

        pages = -(-num // self.metas_per_page)

        paged_metas = zip(count(start=1), get_single_page_metas())

        for page, metas in takewhile(eager_for_metas, paged_metas):
            params['before'] = metas[-1]['id']
            metas = filter_metas(metas)
            if len(metas) == 0:
                return
            yield metas

    def get_post(self, post_id, addition=None, params=None):
        req = self.client.get(route.post(post_id, addition=addition), params=params)
        return req

    get_post_links = partialmethod(get_post, addition='links')

    def imap(self, *args, **kwargs):
        return self.client.imap(*args, **kwargs)

    def get_json(self, *args, **kwargs):
        return self.client.get_json(*args, **kwargs)


class Route():

    host = 'https://www.dcard.tw/'

    def forums(self):
        return Route.host + '_api/forums'

    def posts_meta(self, forum):
        return Route.host + '_api/forums/{forum}/posts'.format(forum=forum)

    def post(self, post_id, addition=None):
        base = Route.host + '_api/posts/{id}'.format(id=post_id)
        if addition:
            return base + '/' + addition
        return base

    post_links = partialmethod(post, addition='links')
    post_comments = partialmethod(post, addition='comments')


route = Route()
api = Api()



from __future__ import absolute_import, unicode_literals

import os
import re
import logging
import contextlib
from multiprocessing.dummy import Pool

from dcard.api import api
from dcard.utils import flatten_lists

logger = logging.getLogger(__name__)


class Downloader:

    def __init__(
            self, download_folder=None, subfolder_pattern=None, flatten=False):
        self.resources_folder = download_folder or './downloads'
        self.subfolder_pattern = subfolder_pattern or '({id}) {folder_name}'
        self.flatten = flatten
        self.resource_bundles = None

    def download(self):
        logger.info('[Downloader] takes hand')

        self.mkdir(self.resources_folder)

        tasks = [
            (self.get_filepath(meta, url), url)
            for meta, urls in self.resource_bundles
            for url in urls
        ]

        with contextlib.closing(Pool(8)) as pool:
            results = pool.map(self.downloading, tasks)

        status = [ok for ok, _ in results]
        fails = [src for ok, src in results if not ok]

        logger.info('[Downloader] download %d items (Total: %d)!',
                    sum(status), len(status))

        return sum(status), fails

    def get_filepath(self, meta, url):
        folder = self.get_folder_fullname(meta)
        filepath = '{folder_name}{separator}{file_name}'.format(
            folder_name=folder,
            file_name=os.path.basename(url),
            separator='-' if self.flatten else '/'
        )

        if not self.flatten:
            self.mkdir(folder)

        return filepath

    def get_folder_fullname(self, meta):
        safe_title = ''.join([
            char
            for char in re.sub('[\?\\\\/><:"|\*\.]', '', meta['title']).strip()
            if ord(char) > 0x1f
        ])

        meta['folder_name'] = safe_title
        folder = self.subfolder_pattern.format(**meta)
        return self.resources_folder + '/' + folder

    @staticmethod
    def mkdir(path):
        if not os.path.exists(path):
            os.makedirs(path)

    @staticmethod
    def save_file(resp, path):
        with open(path, 'wb') as stream:
            for chunk in resp.iter_content(chunk_size=1024):
                stream.write(chunk)

    @classmethod
    def downloading(cls, task):
        filepath, src = task

        if os.path.exists(filepath):
            return True, src
        response = api.client.get_stream(src)

        if response.ok:
            cls.save_file(response, filepath)
        return (response.ok, src)


class ContentParser:

    reg_images = re.compile('http[s]?://\S+\.(?:jpg|png|gif)')
    reg_imgur = re.compile('http[s]?://imgur.com/(\w+)')
    reg_imgur_file = re.compile('http[s]?://i.imgur.com/\w+\.(?:jpg|png|gif)')

    pattern_imgur_file = 'http://i.imgur.com/{img_hash}.jpg'

    def __init__(self):
        self.posts = None

    def parse(self):

        def parsed_post(posts):
            for post in posts:
                article = post.get('content') or ''
                comments = post.get('comments') or []

                imgs = self.parse_images(article) + flatten_lists([
                    self.parse_images(cmt.get('content', ''))
                    for cmt in comments
                ])

                if len(imgs):
                    yield post, imgs

        logger.info('[ContentParser] takes hand')

        results, resource_count = [], 0
        for post, imgs in parsed_post(self.posts):
            del post['content']
            resource_count += len(imgs)
            results.append((post, imgs))

        logger.info('[ContentParser] collects %d resources', resource_count)

        return results

    @classmethod
    def parse_images(cls, raw):
        imgur_files = []
        imgur_files += cls.reg_imgur_file.findall(raw)
        imgur_files += [
            cls.pattern_imgur_file.format(img_hash=r)
            for r in cls.reg_imgur.findall(raw)
        ]
        return imgur_files



# -*- coding: utf-8 -*-
from __future__ import unicode_literals, absolute_import
import logging

from dcard.api import api
from dcard.posts import Post
from dcard.utils import flatten_lists

__all__ = ['Dcard']

logger = logging.getLogger(__name__)


class Dcard:

    def __init__(self):
        self.forums = Forum()
        self.posts = Post()


class Forum:

    infinite_page = -1

    def __init__(self, name=None):
        self.name = name
        self.api = api

    def __call__(self, name):
        self.name = name
        return self

    def get_metas(
            self, num=30, sort='new', before=None, timebound=None,
            callback=None):
        logger.info('<%s> 開始取得看板內文章資訊', self.name)

        paged_metas = self.api.get_metas(
            self.name, sort, num, before, timebound)

        buff = flatten_lists(metas for metas in paged_metas)
        results = callback(buff) if callback else buff

        logger.info('<%s> 資訊蒐集完成，共%d筆', self.name, len(buff))
        return results

    def get(self, no_school=False):
        if no_school:
            return self.api.get_general_forums()
        return self.api.get_all_forums()



# -*- coding: utf-8 -*-
from __future__ import unicode_literals, absolute_import

import time
import argparse

import dcard
from dcard.dcard import Dcard

parser = argparse.ArgumentParser()
parser.add_argument(
    'mode', help='download / meta mode')
parser.add_argument(
    '-f', '--forum', help='Specific which forum')
parser.add_argument(
    '-n', '--number', type=int, help='Scan through how many posts')
parser.add_argument(
    '-b', '--before', type=int, help='Scan through before specified post ID')
parser.add_argument(
    '-likes', '--likes_threshold', type=int, help='Specific minimum like counts')
parser.add_argument(
    '-o', '--output', type=str, help='Specific folder to store the resources')
parser.add_argument(
    '-F', '--flatten', action='store_true', help='Option for flattening folders')
parser.add_argument(
    '-v', '--verbose', action='store_true', help='Logging verbose information')
parser.add_argument(
    '-V', '--version', action='version', version=dcard.__version__)


def main(args=None):
    args = args or parser.parse_args()
    if args.verbose:
        dcard.add_handles_on_logger()
    if args.mode == 'download':
        if not (args.forum or args.number):
            parser.error('No action requested, add --forum or --number')
        download(args)


def download(args):

    def metas_filter(metas):
        return [meta for meta in metas if meta['likeCount'] >= likes_thesh]

    likes_thesh = args.likes_threshold if args.likes_threshold else 0

    dcard = Dcard()

    start_time = time.time()

    metas = dcard \
        .forums(args.forum) \
        .get_metas(num=args.number, before=args.before, callback=metas_filter)
    posts = dcard.posts(metas).get(comments=False, links=False)

    if args.flatten:
        posts.downloader.subfolder_pattern = '[{likeCount}推] {id}-{folder_name}'
        posts.downloader.flatten = True
    if args.output:
        posts.downloader.resources_folder = args.output

    resources = posts.parse_resources()
    cnt, fails = posts.download(resources)

    print('成功下載 %d items！' % cnt
          if not fails else '出了點錯下載不完全喔')
    print('Finish in {0:.5f} sec(s).'.format(time.time() - start_time))

    return cnt, fails



"""
prequests
~~~~~~~~~

An easy copy of `https://github.com/leVirve/prequests` for convenient.

This module contains an asynchronous replica of ``requests.api``, powered
by futures. All API methods return a ``Request`` instance (as opposed to
``Response``). A list of requests can be sent with ``map()``.
"""

import traceback
import contextlib
from functools import partial
from multiprocessing.dummy import Pool

try:
    from requests import Session
except ImportError:
    raise RuntimeError('requests is required for prequests')


__all__ = (
    'map', 'imap', 'imap_unordered',
    'get', 'options', 'head', 'post', 'put', 'patch', 'delete', 'request'
)


class AsyncRequest(object):
    """ Asynchronous request.

    Accept same parameters as ``Session.request`` and some additional:

    :param session: Session which will do request
    :param callback: Callback called on response.
    Same as passing ``hooks={'response': callback}``
    """
    def __init__(self, method, url, **kwargs):
        #: Request method
        self.method = method
        #: URL to request
        self.url = url
        #: Associated ``Session``
        self.session = kwargs.pop('session', None)
        if self.session is None:
            self.session = Session()

        callback = kwargs.pop('callback', None)
        if callback:
            kwargs['hooks'] = {'response': callback}

        #: The rest arguments for ``Session.request``
        self.kwargs = kwargs
        #: Resulting ``Response``
        self.response = None

    def send(self, **kwargs):
        """
        Prepares request based on parameter passed to constructor and optional ``kwargs```.
        Then sends request and saves response to :attr:`response`

        :returns: ``Response``
        """
        merged_kwargs = {}
        merged_kwargs.update(self.kwargs)
        merged_kwargs.update(kwargs)
        try:
            self.response = self.session.request(
                                self.method, self.url, **merged_kwargs)
        except Exception as e:
            self.exception = e
            self.traceback = traceback.format_exc()
        return self


def send(r, stream=False):
    """Just sends the request using its send method and returns its response.  """
    r.send(stream=stream)
    return r.response


# Shortcuts for creating AsyncRequest with appropriate HTTP method
get = partial(AsyncRequest, 'GET')
options = partial(AsyncRequest, 'OPTIONS')
head = partial(AsyncRequest, 'HEAD')
post = partial(AsyncRequest, 'POST')
put = partial(AsyncRequest, 'PUT')
patch = partial(AsyncRequest, 'PATCH')
delete = partial(AsyncRequest, 'DELETE')


# synonym
def request(method, url, **kwargs):
    return AsyncRequest(method, url, **kwargs)


def map(requests, stream=True, pool=None, size=1, exception_handler=None):
    """Concurrently converts a list of Requests to Responses.

    :param requests: a collection of Request objects.
    :param stream: If False, the content will not be downloaded immediately.
    :param size: Specifies the number of workers to run at a time. If 1, no parallel processing.
    :param exception_handler: Callback function, called when exception occured. Params: Request, Exception
    """

    pool = pool if pool else Pool(size)
    requests = list(requests)

    requests = pool.map(send, requests)

    ret = []
    for request in requests:
        if request.response is not None:
            ret.append(request.response)
        elif exception_handler and hasattr(request, 'exception'):
            ret.append(exception_handler(request, request.exception))
        else:
            ret.append(None)

    if not pool:
        pool.close()

    return ret


def imap(requests, stream=True, pool=None, size=2, exception_handler=None):
    """Concurrently converts a generator object of Requests to
    a generator of Responses.

    :param requests: a generator of Request objects.
    :param stream: If False, the content will not be downloaded immediately.
    :param size: Specifies the number of requests to make at a time. default is 2
    :param exception_handler: Callback function, called when exception occured. Params: Request, Exception
    """

    def send(r):
        return r.send(stream=stream)

    pool = pool if pool else Pool(size)

    for request in pool.imap(send, requests):
        if request.response is not None:
            yield request.response
        elif exception_handler:
            exception_handler(request, request.exception)

    if not pool:
        pool.close()


def imap_unordered(requests, stream=True, pool=None, size=2, exception_handler=None):
    """Concurrently converts a generator object of Requests to
    a generator of Responses.

    :param requests: a generator of Request objects.
    :param stream: If False, the content will not be downloaded immediately.
    :param size: Specifies the number of requests to make at a time. default is 2
    :param exception_handler: Callback function, called when exception occured. Params: Request, Exception
    """

    def send(r):
        return r.send(stream=stream)

    pool = pool if pool else Pool(size)

    with contextlib.closing(Pool(size)) as pool:
        for request in pool.imap_unordered(send, requests):
            if request.response is not None:
                yield request.response
            elif exception_handler:
                exception_handler(request, request.exception)

    if not pool:
        pool.close()



import codecs
from setuptools import setup


def readme():
    with codecs.open('README.rst', 'r', 'utf-8') as f:
        return f.read()


def version():
    with open('dcard/__init__.py') as f:
        for line in f:
            if line.startswith('__version__'):
                return line.replace("'", '').split()[-1]


setup(
    name='dcard-spider',
    version=version(),
    url='http://github.com/leVirve/dcard-spider',
    description='A spider for Dcard through its newest API.',
    long_description=readme(),
    author='Salas leVirve',
    author_email='gae.m.project@gmail.com',
    license='MIT',
    platforms='any',
    packages=['dcard'],
    zip_safe=False,
    keywords='Dcard crawler spider',
    entry_points={
        'console_scripts': ['dcard=dcard.cli:main'],
    },
    install_requires=[
        'six',
        'requests'
    ],
    dependency_links=[
        'https://github.com/leVirve/prequests/tarball/master#egg=prequests-0.1.1'
    ],
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Intended Audience :: Customer Service',
        'Intended Audience :: System Administrators',
        'License :: OSI Approved :: MIT License',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Software Development :: Libraries',
        'Topic :: Terminals',
        'Topic :: Text Processing',
        'Topic :: Utilities',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6'
    ]
)
